@@ -45,6 +45,7 @@
 #include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -61,6 +62,10 @@
 #include <sys/wait.h>
 #include <unistd.h>
 
+#ifdef WITH_SYSTEMD
+#include <systemd/sd-daemon.h>
+#endif
+
 #ifdef HAVE_SIGHANDLER_T
 #define SIGHANDLER_T sighandler_t
 #elif HAVE_SIG_T
@@ -142,13 +147,17 @@ server_new (const char **tokens, size_t n_tokens, const char *provider,
 	if (strncmp (tokens[0], "pkcs11:", 7) == 0) {
 		if (server->provider) {
 			server->module = p11_kit_module_load (server->provider, 0);
-			if (server->module == NULL)
+			if (server->module == NULL) {
+				free (server);
 				return NULL;
+			}
 		}
 	} else {
 		server->module = p11_kit_module_load (tokens[0], 0);
-		if (server->module == NULL)
+		if (server->module == NULL) {
+			free (server);
 			return NULL;
+		}
 	}
 #endif /* OS_WIN32 */
 
@@ -161,6 +170,10 @@ static bool need_children_cleanup = false;
 static bool terminate = false;
 static unsigned children_avail = 0;
 static bool quiet = false;
+static bool csh = false;
+
+#define P11_KIT_SERVER_ADDRESS_ENV "P11_KIT_SERVER_ADDRESS"
+#define P11_KIT_SERVER_PID_ENV "P11_KIT_SERVER_PID"
 
 static SIGHANDLER_T
 ocsignal (int signum, SIGHANDLER_T handler)
@@ -268,13 +281,22 @@ create_socket (const char *address,
 	umask (066);
 	rc = bind (sd, (struct sockaddr *)&sa, SUN_LEN (&sa));
 	if (rc == -1) {
-		p11_message_err (errno, "could not create socket %s", socket_file);
+		close (sd);
+		p11_message_err (errno, "could not bind socket %s", socket_file);
 		return -1;
 	}
 
+	rc = listen (sd, 1024);
+	if (rc == -1) {
+		close (sd);
+		p11_message_err (errno, "could not listen to socket %s", socket_file);
+		return 1;
+	}
+
 	if (uid != -1 && gid != -1) {
 		rc = chown (socket_file, uid, gid);
 		if (rc == -1) {
+			close (sd);
 			p11_message_err (errno, "could not chown socket %s", socket_file);
 			return -1;
 		}
@@ -316,12 +338,42 @@ check_credentials (int fd,
 	return true;
 }
 
+static bool
+print_environment (pid_t pid, const char *socket_path, bool csh)
+{
+	char *path, *address;
+	int rc;
+
+	path = p11_path_encode (socket_path);
+	rc = asprintf (&address, "unix:path=%s", path);
+	free (path);
+	if (rc < 0)
+		return false;
+	if (csh) {
+		printf ("setenv %s %s;\n",
+			P11_KIT_SERVER_ADDRESS_ENV,
+			address);
+		printf ("setenv %s %d;\n",
+			P11_KIT_SERVER_PID_ENV,
+			pid);
+	} else {
+		printf ("%s=%s; export %s;\n",
+			P11_KIT_SERVER_ADDRESS_ENV, address,
+			P11_KIT_SERVER_ADDRESS_ENV);
+		printf ("%s=%d; export %s;\n",
+			P11_KIT_SERVER_PID_ENV, pid,
+			P11_KIT_SERVER_PID_ENV);
+	}
+	free (address);
+	return true;
+}
+
 static int
 server_loop (Server *server,
 	     bool foreground,
 	     struct timespec *timeout)
 {
-	int ret = 1, rc;
+	int ret;
 	int cfd;
 	pid_t pid;
 	socklen_t sa_len;
@@ -342,21 +394,21 @@ server_loop (Server *server,
 	ocsignal (SIGTERM, handle_term);
 	ocsignal (SIGINT, handle_term);
 
-	server->socket = create_socket (server->socket_name, server->uid, server->gid);
-	if (server->socket == -1)
-		return 1;
-
 	/* run as daemon */
 	if (!foreground) {
 		pid = fork ();
-		switch (pid) {
-		case -1:
+		if (pid == -1) {
 			p11_message_err (errno, "could not fork() to daemonize");
 			return 1;
-		case 0:
-			break;
-		default:
-			_exit (0);
+		}
+		if (pid == 0) {
+			close (STDIN_FILENO);
+			close (STDOUT_FILENO);
+		}
+		if (pid != 0) {
+			if (!print_environment (pid, server->socket_name, csh))
+				return 1;
+			exit (0);
 		}
 		if (setsid () == -1) {
 			p11_message_err (errno, "could not create a new session");
@@ -364,21 +416,29 @@ server_loop (Server *server,
 		}
 	}
 
-	rc = listen (server->socket, 1024);
-	if (rc == -1) {
-		p11_message_err (errno, "could not listen to socket %s", server->socket_name);
+#ifdef WITH_SYSTEMD
+	ret = sd_listen_fds (0);
+	if (ret > 1) {
+		p11_message ("too many file descriptors received");
 		return 1;
+	} else if (ret == 1) {
+		server->socket = SD_LISTEN_FDS_START + 0;
+	} else
+#endif
+	{
+		server->socket = create_socket (server->socket_name, server->uid, server->gid);
+		if (server->socket == -1)
+			return 1;
 	}
 
 	sigprocmask (SIG_BLOCK, &blockset, NULL);
 
-	if (!quiet) {
-		char *path;
-
-		path = p11_path_encode (server->socket_name);
-		printf ("P11_KIT_SERVER_ADDRESS=unix:path=%s\n", path);
-		free (path);
-		printf ("P11_KIT_SERVER_PID=%d\n", getpid ());
+	/* for testing purposes, even when started in foreground,
+	 * print the envvars */
+	if (foreground) {
+		if (!print_environment (getpid (), server->socket_name, csh))
+			return 1;
+		fflush (stdout);
 	}
 
 	/* accept connections */
@@ -397,7 +457,8 @@ server_loop (Server *server,
 		if (ret == -1 && errno == EINTR)
 			continue;
 
-		if (ret == 0 && children_avail == 0) { /* timeout */
+		/* timeout */
+		if (ret == 0 && children_avail == 0 && timeout != NULL) {
 			p11_message ("no connections to %s for %lu secs, exiting", server->socket_name, timeout->tv_sec);
 			break;
 		}
@@ -442,7 +503,7 @@ server_loop (Server *server,
 				}
 
 				n_args = 0;
-				args[n_args] = "p11-kit-remote";
+				args[n_args] = P11_KIT_REMOTE;
 				n_args++;
 
 				if (server->provider) {
@@ -485,6 +546,8 @@ main (int argc,
 	const struct passwd *pwd;
 	const struct group *grp;
 	bool foreground = false;
+	bool csh_opt = false;
+	bool kill_opt = false;
 	struct timespec *timeout = NULL, ts;
 	char *name = NULL;
 	char *provider = NULL;
@@ -502,7 +565,10 @@ main (int argc,
 		opt_foreground = 'f',
 		opt_timeout = 't',
 		opt_name = 'n',
-		opt_provider = 'p'
+		opt_provider = 'p',
+		opt_kill = 'k',
+		opt_csh = 'c',
+		opt_sh = 's'
 	};
 
 	struct option options[] = {
@@ -517,6 +583,9 @@ main (int argc,
 		{ "timeout", required_argument, NULL, opt_timeout },
 		{ "name", required_argument, NULL, opt_name },
 		{ "provider", required_argument, NULL, opt_provider },
+		{ "kill", no_argument, NULL, opt_kill },
+		{ "csh", no_argument, NULL, opt_csh },
+		{ "sh", no_argument, NULL, opt_sh },
 		{ 0 },
 	};
 
@@ -530,6 +599,9 @@ main (int argc,
 		{ opt_timeout, "exit if no connection until the given timeout" },
 		{ opt_name, "specify name of the socket (default: pkcs11-<pid>)" },
 		{ opt_provider, "specify the module to use" },
+		{ opt_kill, "terminate the running server" },
+		{ opt_csh, "generate C-shell commands on stdout" },
+		{ opt_sh, "generate Bourne shell commands on stdout" },
 		{ 0 },
 	};
 
@@ -587,6 +659,17 @@ main (int argc,
 		case opt_provider:
 			provider = optarg;
 			break;
+		case opt_kill:
+			kill_opt = true;
+			break;
+		case opt_csh:
+			csh = true;
+			csh_opt = true;
+			break;
+		case opt_sh:
+			csh = false;
+			csh_opt = true;
+			break;
 		case opt_help:
 		case '?':
 			p11_tool_usage (usages, options);
@@ -600,11 +683,54 @@ main (int argc,
 	argc -= optind;
 	argv += optind;
 
-	if (argc < 1) {
+	if (argc < 1 && !kill_opt) {
 		p11_tool_usage (usages, options);
 		return 2;
 	}
 
+	if (!csh_opt) {
+		const char *shell = secure_getenv ("SHELL");
+		size_t len;
+		if (shell != NULL && (len = strlen (shell)) > 2 &&
+		    strncmp (shell + len - 3, "csh", 3) == 0)
+			csh = true;
+	}
+
+	if (kill_opt) {
+		const char *pidstr = secure_getenv (P11_KIT_SERVER_PID_ENV);
+		char *endptr;
+		long pidval;
+
+		if (pidstr == NULL) {
+			fprintf (stderr, "%s not set, cannot kill server",
+				 P11_KIT_SERVER_PID_ENV);
+			exit (1);
+		}
+		pidval = strtol (pidstr, &endptr, 10);
+		if (errno == ERANGE &&
+		    (pidval == LONG_MAX || pidval == LONG_MIN)) {
+			perror ("strtol");
+			exit (1);
+		}
+		if (kill ((pid_t) pidval, SIGTERM) == -1) {
+			perror ("kill");
+			exit (1);
+		}
+
+		if (csh) {
+			printf ("unsetenv %s;\n",
+				P11_KIT_SERVER_ADDRESS_ENV);
+			printf ("unsetenv %s;\n",
+				P11_KIT_SERVER_PID_ENV);
+		} else {
+			printf ("unset %s;\n",
+				P11_KIT_SERVER_ADDRESS_ENV);
+			printf ("unset %s;\n",
+				P11_KIT_SERVER_PID_ENV);
+		}
+		exit (0);
+	}
+
 	if (run_as_gid != -1) {
 		if (setgid (run_as_gid) == -1) {
 			p11_message_err (errno, "cannot set gid to %u", (unsigned)run_as_gid);
