@@ -187,7 +187,10 @@ p11_builder_new (int flags)
 	return_val_if_fail (builder != NULL, NULL);
 
 	builder->asn1_cache = p11_asn1_cache_new ();
-	return_val_if_fail (builder->asn1_cache, NULL);
+	if (builder->asn1_cache == NULL) {
+		p11_builder_free (builder);
+		return_val_if_reached (NULL);
+	}
 	builder->asn1_defs = p11_asn1_cache_defs (builder->asn1_cache);
 
 	builder->flags = flags;
@@ -472,9 +475,9 @@ calc_date (node_asn *node,
 		return_val_if_fail (year >= 0, false);
 
 		century = century_for_two_digit_year (year);
-		return_val_if_fail (century >= 0, false);
+		return_val_if_fail (century >= 0 && century <= 9900, false);
 
-		snprintf ((char *)date->year, 3, "%02d", century);
+		snprintf ((char *)date->year, 3, "%02d", century / 100);
 		memcpy (((char *)date) + 2, buf, 6);
 
 	} else {
