@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013, Red Hat Inc.
+ * Copyright (c) 2019, Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -48,6 +48,10 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
+#include <errno.h>
+#include <limits.h>
+
+time_t _p11_extract_jks_timestamp = 0;
 
 static void
 encode_msb_short (unsigned char *data,
@@ -157,7 +161,7 @@ convert_alias (const char *input,
 	for (i = 0; i < length; i++) {
 		ch = input[i];
 		if (!isspace (ch) && (ch & 0x80) == 0) {
-			ch = tolower (ch);
+			ch = p11_ascii_tolower (ch);
 			p11_buffer_add (buf, &ch, 1);
 		}
 	}
@@ -187,6 +191,7 @@ add_alias (p11_buffer *buffer,
 	}
 
 	convert_alias (input, input_len, &buf);
+	length = buf.len;
 
 	for (i = 0; i < INT32_MAX; i++) {
 		if (i > 0) {
@@ -203,7 +208,7 @@ add_alias (p11_buffer *buffer,
 			return true;
 		}
 
-		p11_buffer_reset (&buf, 0);
+		buf.len = length;
 	}
 
 	return false;
@@ -245,7 +250,38 @@ prepare_jks_buffer (p11_enumerate *ex,
 	 * when this was this certificate was added to the keystore, however
 	 * we don't have that information. Java uses time in milliseconds
 	 */
-	now = time (NULL);
+	{
+		char *source_date_epoch;
+		source_date_epoch = secure_getenv ("SOURCE_DATE_EPOCH");
+		if (source_date_epoch) {
+			unsigned long long epoch;
+			char *endptr;
+			errno = 0;
+			epoch = strtoull (source_date_epoch, &endptr, 10);
+			if ((errno == ERANGE && (epoch == ULLONG_MAX || epoch == 0))
+			    || (errno != 0 && epoch == 0)) {
+				p11_message_err (errno, "Environment variable $SOURCE_DATE_EPOCH: strtoull");
+				return false;
+			}
+			if (endptr == source_date_epoch) {
+				fprintf (stderr, "Environment variable $SOURCE_DATE_EPOCH: No digits were found: %s\n", endptr);
+				return false;
+			}
+			if (*endptr != '\0') {
+				fprintf (stderr, "Environment variable $SOURCE_DATE_EPOCH: Trailing garbage: %s\n", endptr);
+				return false;
+			}
+			if (epoch > ULONG_MAX) {
+				fprintf (stderr, "Environment variable $SOURCE_DATE_EPOCH: value must be smaller than or equal to %lu but was found to be: %llu \n", ULONG_MAX, epoch);
+				return false;
+			}
+			now = epoch;
+		} else if (_p11_extract_jks_timestamp)
+			now = _p11_extract_jks_timestamp;
+		else
+			now = time (NULL);
+	}
+
 	return_val_if_fail (now > 0, false);
 	now *= 1000; /* seconds to milliseconds */
 
