@@ -119,6 +119,8 @@ teardown (void *unused)
 	rv = test.module->C_Finalize (NULL);
 	assert (rv == CKR_OK);
 
+	if (test.directory)
+		p11_test_directory_delete (test.directory);
 	free (test.directory);
 
 	memset (&test, 0, sizeof (test));
@@ -315,8 +317,8 @@ test_get_token_info (void)
 
 	memset (&args, 0, sizeof (args));
 	args.pReserved = "paths='" \
-		SYSCONFDIR "/trust/input" P11_PATH_SEP \
-		DATA_DIR "/trust/fixtures/blah" P11_PATH_SEP \
+		P11_SYSTEM_TRUST_PREFIX "/trust/input" P11_PATH_SEP \
+		P11_DEFAULT_TRUST_PREFIX "/trust/fixtures/blah" P11_PATH_SEP \
 		"/some/other/path/the-basename'";
 	args.flags = CKF_OS_LOCKING_OK;
 
@@ -623,12 +625,54 @@ test_find_certificates (void)
 	CK_ULONG i;
 
 	count = find_objects (match, sessions, objects, 16);
-	assert_num_eq (8, count);
+	assert_num_eq (9, count);
 
 	for (i = 0; i < count; i++)
 		check_certificate (sessions[i], objects[i]);
 }
 
+static void
+test_find_extensions (void)
+{
+	CK_OBJECT_CLASS klass = CKO_X_CERTIFICATE_EXTENSION;
+	unsigned char spki[] = {
+		0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a,
+		0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01,
+		0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81,
+		0x89, 0x02, 0x81, 0x81, 0x00, 0xd1, 0xb5, 0x36,
+		0xa3, 0x89, 0xee, 0xaa, 0x80, 0x2f, 0x53, 0xfd,
+		0x12, 0x75, 0x3e, 0xf3, 0x7a, 0x9e, 0xd6, 0xaf,
+		0xfa, 0xbc, 0x1c, 0x60, 0x10, 0x4b, 0x26, 0x81,
+		0x13, 0x1a, 0x59, 0xe3, 0xfe, 0x45, 0x6c, 0x38,
+		0x04, 0x39, 0x27, 0x46, 0x57, 0xfd, 0xd5, 0xbc,
+		0x8d, 0x8a, 0x10, 0xb6, 0x3b, 0xd4, 0x0a, 0x81,
+		0x5a, 0x45, 0x2f, 0xec, 0x3e, 0x81, 0xf0, 0xd9,
+		0x4e, 0x4f, 0x97, 0xdf, 0x4b, 0x32, 0x0f, 0x08,
+		0xb1, 0x26, 0xa2, 0xbd, 0x69, 0x61, 0x5d, 0x66,
+		0x39, 0x63, 0x2f, 0x10, 0x70, 0x35, 0xfb, 0x07,
+		0x85, 0x0a, 0xff, 0x57, 0x12, 0xc1, 0xf4, 0x83,
+		0x1d, 0xf9, 0xc6, 0xd3, 0xa4, 0xb6, 0x70, 0x2b,
+		0x80, 0xa1, 0x40, 0x7f, 0x48, 0x4e, 0xd9, 0xad,
+		0xeb, 0x80, 0xcc, 0xcf, 0x92, 0xc1, 0xd1, 0x83,
+		0x64, 0x01, 0x23, 0x47, 0x8e, 0xbd, 0x31, 0x98,
+		0x05, 0x6b, 0x6b, 0x7c, 0x37, 0x02, 0x03, 0x01,
+		0x00, 0x01
+	};
+
+	CK_ATTRIBUTE match[] = {
+		{ CKA_CLASS, &klass, sizeof (klass) },
+		{ CKA_PUBLIC_KEY_INFO, spki, sizeof (spki) },
+		{ CKA_INVALID, }
+	};
+
+	CK_OBJECT_HANDLE objects[16];
+	CK_SESSION_HANDLE sessions[16];
+	CK_ULONG count;
+
+	count = find_objects (match, sessions, objects, 16);
+	assert_num_eq (1, count);
+}
+
 static void
 test_find_builtin (void)
 {
@@ -1175,6 +1219,68 @@ test_modify_and_write (void)
 	test_check_attrs (expected, parsed->elem[0]);
 }
 
+static void
+test_token_write_protected (void)
+{
+	CK_C_INITIALIZE_ARGS args;
+	CK_FUNCTION_LIST *module;
+	CK_SLOT_ID slots[NUM_SLOTS];
+	CK_TOKEN_INFO info;
+	char label[32];
+	CK_ULONG count;
+	CK_RV rv;
+	int i;
+
+	/* These are the paths passed in in setup() */
+	const char *labels[] = {
+		"System Trust",
+		"Default Trust",
+		"the-basename",
+	};
+
+	/* This is the entry point of the trust module, linked to this test */
+	rv = C_GetFunctionList (&module);
+	assert (rv == CKR_OK);
+
+	memset (&args, 0, sizeof (args));
+	args.pReserved = "paths='" \
+		P11_SYSTEM_TRUST_PREFIX "/trust/input" P11_PATH_SEP \
+		P11_DEFAULT_TRUST_PREFIX "/trust/fixtures/blah" P11_PATH_SEP \
+		"/some/other/path/the-basename'";
+	args.flags = CKF_OS_LOCKING_OK;
+
+	rv = module->C_Initialize (&args);
+	assert (rv == CKR_OK);
+
+	count = NUM_SLOTS;
+	rv = module->C_GetSlotList (CK_TRUE, slots, &count);
+	assert (rv == CKR_OK);
+	assert (count == NUM_SLOTS);
+
+	for (i = 0; i < NUM_SLOTS; i++) {
+		rv = module->C_GetTokenInfo (slots[i], &info);
+		assert_num_eq (CKR_OK, rv);
+
+		memset (label, ' ', sizeof (label));
+		memcpy (label, labels[i], strlen (labels[i]));
+		assert (memcmp (info.label, label, sizeof (label)) == 0);
+
+		switch (i) {
+		case 0:
+			assert_num_cmp (0, ==, info.flags & CKF_WRITE_PROTECTED);
+			break;
+		case 1:
+			assert_num_cmp (0, !=, info.flags & CKF_WRITE_PROTECTED);
+			break;
+		default:
+			break;
+		}
+	}
+
+	rv = module->C_Finalize (NULL);
+	assert_num_eq (CKR_OK, rv);
+}
+
 int
 main (int argc,
       char *argv[])
@@ -1194,6 +1300,7 @@ main (int argc,
 	p11_test (test_get_session_info, "/module/get_session_info");
 	p11_test (test_close_all_sessions, "/module/close_all_sessions");
 	p11_test (test_find_certificates, "/module/find_certificates");
+	p11_test (test_find_extensions, "/module/find_extensions");
 	p11_test (test_find_builtin, "/module/find_builtin");
 	p11_test (test_lookup_invalid, "/module/lookup_invalid");
 	p11_test (test_remove_token, "/module/remove_token");
@@ -1214,5 +1321,8 @@ main (int argc,
 	p11_test (test_create_and_write, "/module/create-and-write");
 	p11_test (test_modify_and_write, "/module/modify-and-write");
 
+	p11_fixture (NULL, NULL);
+	p11_test (test_token_write_protected, "/module/token-write-protected");
+
 	return p11_test_run (argc, argv);
 }
