@@ -196,14 +196,14 @@ loader_load_file (p11_token *token,
 	default:
 		p11_debug ("failed to parse: %s", filename);
 		loader_gone_file (token, filename);
-		return 0;
+		return -1;
 	}
 
 	/* Update each parsed object with the origin */
 	parsed = p11_parser_parsed (token->parser);
 	for (i = 0; i < parsed->num; i++) {
 		parsed->elem[i] = p11_attrs_build (parsed->elem[i], origin, NULL);
-		return_val_if_fail (parsed->elem[i] != NULL, 0);
+		return_val_if_fail (parsed->elem[i] != NULL, -1);
 	}
 
 	p11_index_load (token->index);
@@ -215,7 +215,7 @@ loader_load_file (p11_token *token,
 
 	if (rv != CKR_OK) {
 		p11_message ("couldn't load file into objects: %s", filename);
-		return 0;
+		return -1;
 	}
 
 	loader_was_loaded (token, filename, sb);
@@ -377,16 +377,19 @@ p11_token_load (p11_token *token)
 	int ret;
 
 	ret = loader_load_path (token, token->path, &is_dir);
-	return_val_if_fail (ret >= 0, -1);
+	if (ret < 0)
+		return -1;
 	total += ret;
 
 	if (is_dir) {
 		ret = loader_load_path (token, token->anchors, &is_dir);
-		return_val_if_fail (ret >= 0, -1);
+		if (ret < 0)
+			return -1;
 		total += ret;
 
 		ret = loader_load_path (token, token->blacklist, &is_dir);
-		return_val_if_fail (ret >= 0, -1);
+		if (ret < 0)
+			return -1;
 		total += ret;
 	}
 
@@ -817,7 +820,8 @@ p11_token_free (p11_token *token)
 p11_token *
 p11_token_new (CK_SLOT_ID slot,
                const char *path,
-               const char *label)
+               const char *label,
+               int flags)
 {
 	p11_token *token;
 
@@ -828,7 +832,10 @@ p11_token_new (CK_SLOT_ID slot,
 	return_val_if_fail (token != NULL, NULL);
 
 	token->builder = p11_builder_new (P11_BUILDER_FLAG_TOKEN);
-	return_val_if_fail (token->builder != NULL, NULL);
+	if (token->builder == NULL) {
+		p11_token_free (token);
+		return_val_if_reached (NULL);
+	}
 
 	token->index = p11_index_new (on_index_build,
 	                              on_index_store,
@@ -859,6 +866,12 @@ p11_token_new (CK_SLOT_ID slot,
 
 	token->slot = slot;
 
+	if (flags & P11_TOKEN_FLAG_WRITE_PROTECTED) {
+		token->checked_path = true;
+		token->make_directory = false;
+		token->is_writable = false;
+	}
+
 	load_builtin_objects (token);
 
 	p11_debug ("token: %s: %s", token->label, token->path);
